﻿using System;
using System.Collections.Generic;
using MathNet.Numerics.Distributions;

namespace CustomerBehaviour
{
    class Customer
    {
        private readonly double _distanceMultiplier;
        private readonly double _serviceMultiplier;
        private readonly double _memoryFactor;
        private readonly double _influenceFactor;
        private readonly double _renewAlpha;
        private readonly double _renewBeta;
        private readonly double _freqLambda;
        private readonly double _sevAlpha;
        private readonly double _sevBeta;
        private readonly double _transmissionProb;

        private readonly Random _localRand;
        private double _loss;
        private List<Customer> friendsList;
        private readonly int _custIndex;
        internal readonly double location;

        private readonly int _numInsurers;
        private readonly double[] insDistanceCost;
        internal readonly double[] insServiceBonus;
        private readonly double[] insServiceTotalCount;
        internal readonly double[] insServiceExperienceAvg;
        private readonly double[] insServiceExperienceCount;
        internal readonly double[] insServiceWOMAvg;
        private readonly double[] insServiceWOMCount;
        internal double currentOpinion;

        private double _renewProbability;
        private bool _renewed;
        private double _renewCost;
        internal int chosenInsurer;
        internal double premiumCost;
        internal double renewPremiumCost;
        internal double locationCost;
        internal double renewLocationCost;
        internal double serviceCost;
        internal double renewServiceCost;
        internal double womCost;
        internal double renewWomCost;

        private double _oldCost;

        internal Customer(int custIndex, double setLocation, Random seedRand, double[] insurerLocationList, int oldInsurer, double oldPremium, double distanceMultiplier, double serviceMultiplier, double memoryFactor, double influenceFactor, double renewAlpha, double renewBeta, double freqLambda, double sevAlpha, double sevBeta, double transmissionProb)
        {
            _distanceMultiplier = distanceMultiplier;
            _serviceMultiplier = serviceMultiplier;
            _memoryFactor = memoryFactor;
            _influenceFactor = influenceFactor;
            _renewAlpha = renewAlpha;
            _renewBeta = renewBeta;
            _freqLambda = freqLambda;
            _sevAlpha = sevAlpha;
            _sevBeta = sevBeta;
            _transmissionProb = transmissionProb;

            _localRand = new Random(seedRand.Next());
            _custIndex = custIndex;
            location = setLocation;
            _numInsurers = insurerLocationList.Length;

            insDistanceCost = new double[_numInsurers];
            insServiceBonus = new double[_numInsurers];
            insServiceTotalCount = new double[_numInsurers];
            insServiceExperienceAvg = new double[_numInsurers];
            insServiceExperienceCount = new double[_numInsurers];
            insServiceWOMAvg = new double[_numInsurers];
            insServiceWOMCount = new double[_numInsurers];

            for (int i=0;i<_numInsurers;i++)
            {
                double distCost = Math.Abs(location - insurerLocationList[i]);
                if (distCost > 1) { distCost = 2 - distCost; }
                insDistanceCost[i] = distCost;
            }
            NextSim(oldInsurer, oldPremium);
        }
        internal void NextSim(int oldInsurer, double oldPremium)
        {
            for (int i = 0; i < _numInsurers; i++)
            {
                insServiceBonus[i] = 0;
                insServiceTotalCount[i] =0;
                insServiceExperienceAvg[i] = 0;
                insServiceExperienceCount[i] = 0;
                insServiceWOMAvg[i] = 0;
                insServiceWOMCount[i] = 0;
            }
            chosenInsurer = oldInsurer;

            premiumCost = oldPremium;
            locationCost= _distanceMultiplier * insDistanceCost[chosenInsurer];
            serviceCost = 0;
            womCost = 0;
            _oldCost = oldPremium + locationCost;
            friendsList = new List<Customer>();
        }
        internal void SetFriendsList(List<Customer> networkFriendsList)
        {
            friendsList = networkFriendsList;
        }

        internal bool Renew(double renewPremium)
        {
            renewPremiumCost = renewPremium;
            renewLocationCost = _distanceMultiplier * insDistanceCost[chosenInsurer];
            double totalCount = insServiceExperienceCount[chosenInsurer] * (1.0 - _influenceFactor) + insServiceWOMCount[chosenInsurer] * _influenceFactor;
            renewServiceCost = totalCount == 0 ? 0 : (-_serviceMultiplier) * (1.0 - _influenceFactor) * insServiceExperienceAvg[chosenInsurer] * (insServiceExperienceCount[chosenInsurer] / totalCount);
            renewWomCost= totalCount == 0 ? 0 : (-_serviceMultiplier) * _influenceFactor * insServiceWOMAvg[chosenInsurer] * (insServiceWOMCount[chosenInsurer] / totalCount);
            premiumCost = renewPremiumCost;
            locationCost = renewLocationCost;
            serviceCost = renewServiceCost;
            womCost = renewWomCost;
            _renewCost = renewPremium + renewLocationCost - _serviceMultiplier*insServiceBonus[chosenInsurer];

            double costChange = _renewCost / _oldCost - 1.0;
            _renewProbability = 1.0 / (1.0 + Math.Exp(_renewAlpha * costChange +_renewBeta));
            _renewed = _localRand.NextDouble() <= _renewProbability;
            if (_renewed) { _oldCost = _renewCost; }
            return _renewed;
        }
        internal int ChooseNew(double[] premiumOffers)
        {
            double[] newCosts = new double[_numInsurers];
            int oldInsurer = chosenInsurer;
            for(int i=0; i<_numInsurers;i++)
            {
                if (i == chosenInsurer) { newCosts[i] = _renewCost; }
                else { newCosts[i] = premiumOffers[i] + _distanceMultiplier * insDistanceCost[i] - _serviceMultiplier * insServiceBonus[i]; }
            }
            chosenInsurer = -1;
            _oldCost = newCosts[0]+1;
            for (int i = 0; i < _numInsurers; i++)
            {
                if(newCosts[i]<_oldCost || (newCosts[i]==_oldCost && _localRand.NextDouble()<=0.5))
                {
                    chosenInsurer = i;
                    premiumCost = chosenInsurer == oldInsurer ? renewPremiumCost : premiumOffers[i];
                    locationCost = _distanceMultiplier * insDistanceCost[i];
                    double totalCount = insServiceExperienceCount[i] * (1.0 - _influenceFactor) + insServiceWOMCount[i] * _influenceFactor;
                    serviceCost = totalCount == 0 ? 0 : (-_serviceMultiplier) * (1.0 - _influenceFactor) * insServiceExperienceAvg[i] * (insServiceExperienceCount[i] / totalCount);
                    womCost = totalCount == 0 ? 0 : (-_serviceMultiplier) *  _influenceFactor * insServiceWOMAvg[i] * (insServiceWOMCount[i] / totalCount);
                    _oldCost = newCosts[i];
                }
            }
            if (chosenInsurer == oldInsurer) { _renewed = true; }
            return chosenInsurer;
        }
        internal void UpdateOpinion()
        {
            for (int i = 0; i < _numInsurers; i++)
            {
                double totalCount = insServiceExperienceCount[i] * (1.0 - _influenceFactor) + insServiceWOMCount[i] * _influenceFactor;
                insServiceTotalCount[i] = totalCount;
                insServiceBonus[i] = totalCount==0 ? 0 : ( (1.0 - _influenceFactor) * insServiceExperienceAvg[i]*(insServiceExperienceCount[i]/totalCount) + _influenceFactor * insServiceWOMAvg[i]* (insServiceWOMCount[i] / totalCount));
            }
            currentOpinion = insServiceBonus[chosenInsurer];
        }
        internal Loss GetLoss()
        {
            int freq = Poisson.Sample(_localRand,_freqLambda);
            _loss= freq==0 ? 0 : Gamma.Sample(_localRand, freq*_sevAlpha, _sevBeta);
            return new Loss(freq,_loss);
        }
        internal void UpdateService(int newExperienceCount, int newServiceExperience)
        {
            double oldServiceAvg = insServiceExperienceAvg[chosenInsurer] * _memoryFactor;
            double oldCount = insServiceExperienceCount[chosenInsurer];
            double newCount = oldCount * _memoryFactor + newExperienceCount;
            if(newCount > 0)
            { 
                insServiceExperienceAvg[chosenInsurer] = oldServiceAvg * oldCount / newCount + newServiceExperience / newCount;
                insServiceExperienceCount[chosenInsurer] = newCount;
            }
            for(int i=0;i< _numInsurers;i++)
            {
                if(i!=chosenInsurer)
                {
                    insServiceExperienceCount[i] *= _memoryFactor;
                }
            }
        }
        internal void UpdateWOM()
        {
            for (int i = 0; i<_numInsurers; i++)
            {
                double newInfo = 0;
                int newCount = 0;
                for(int j=0; j<friendsList.Count;j++)
                {
                    if (friendsList[j].insServiceTotalCount[i] > 0)
                    {
                        if (_localRand.NextDouble() <= _transmissionProb)
                        {
                            newInfo += friendsList[j].insServiceBonus[i];
                            newCount++;
                        }
                    }
                }
                this.ReceiveWOM(i, newCount, newInfo);
            }
        }
        internal void ReceiveWOM(int insurerIndex, int newWOMCount, double newWOMInfo)
        {
            double oldWOMAvg = insServiceWOMAvg[insurerIndex] * _memoryFactor;
            double oldCount = insServiceWOMCount[insurerIndex];
            double newCount = oldCount * _memoryFactor + newWOMCount;
            if (newCount > 0)
            { insServiceWOMAvg[insurerIndex] = oldWOMAvg * oldCount / newCount + newWOMInfo / newCount;}
            insServiceWOMCount[insurerIndex] = newCount;
        }
    }
    class Loss
    {
        internal int freq;
        internal double sev;
        internal Loss(int setFreq, double setSev)
        {
            freq=setFreq;
            sev=setSev;
        }
    }
}
