﻿using System;

namespace CustomerBehaviour
{
    class Insurer
    {
        private readonly double _maxServiceSpendPerCustomer;
        private ISpendStrategy _spendStrategy;

        private readonly double _renewMarkup;
        private readonly int _renewMarkupMaxTerm;
        private int[] _customerYears;
        private bool[] _customerBought;

        private readonly Random _localRand;
        private readonly int _insIndex;
        private readonly double _location;
        private readonly Outputs outputs;

        private double _newPremium;
        private double _serviceSpendPerCustomer;
        private double _goodServiceProb;

        private double _totalPremium;
        private double _totalLosses;
        private double _totalProfit;
        internal int soldCount;
        private int _renewCount;
        private int _oldSoldCount;
        private int _serviceCount;
        private int _badServicecount;
        private double _totalCustomerLocation;
        private int _rank;

        internal Insurer(Random seedRand, int insIndex, string outputFolder, double location, double renewMarkup, int renewMarkupMaxTerm, double maxServiceSpendPerCustomer, ISpendStrategy spendStrategy, int oldSoldCount, int[] oldSoldByIndex)
        {
            _localRand = new Random(seedRand.Next());
            _renewMarkup = renewMarkup;
            _renewMarkupMaxTerm = renewMarkupMaxTerm;
            _customerBought = new bool[oldSoldByIndex.Length];
            _customerYears = oldSoldByIndex;
            _maxServiceSpendPerCustomer = maxServiceSpendPerCustomer;
            _spendStrategy = spendStrategy;
            _insIndex = insIndex;
            _location = location;
            _oldSoldCount = oldSoldCount;

            outputs = new Outputs(outputFolder);
            outputs.AddOutput("Insurer " + _insIndex + "_New Premium");
            outputs.AddOutput("Insurer " + _insIndex + "_Num New Sales");
            outputs.AddOutput("Insurer " + _insIndex + "_Renewal Rate");
            outputs.AddOutput("Insurer " + _insIndex + "_Total Sales");
            outputs.AddOutput("Insurer " + _insIndex + "_Avg Customer Location");
            outputs.AddOutput("Insurer " + _insIndex + "_Total Service Count");
            outputs.AddOutput("Insurer " + _insIndex + "_Bad Service Count");
            outputs.AddOutput("Insurer " + _insIndex + "_Service Spend Per Customer");
            outputs.AddOutput("Insurer " + _insIndex + "_Profit");
        }
        internal void NextSim(int oldSoldCount, int[] oldSoldByIndex, ISpendStrategy spendStrategy)
        {
            _oldSoldCount = oldSoldCount;
            _customerBought = new bool[oldSoldByIndex.Length];
            _customerYears = oldSoldByIndex;
            outputs.NextSim();
            _totalLosses = 0;
            _totalPremium = 0;
            soldCount = 0;
            _renewCount = 0;
            _serviceCount = 0;
            _badServicecount = 0;
            _totalCustomerLocation = 0;
            _spendStrategy = spendStrategy;
        }
        internal void NextTimestep()
        {
            outputs.SetValues(new double[] { _newPremium, soldCount-_renewCount,_oldSoldCount==0 ? 0: _renewCount/(double)_oldSoldCount, soldCount,
            _totalCustomerLocation/(double)soldCount, _serviceCount, _badServicecount, _serviceSpendPerCustomer, _totalProfit});
            outputs.WriteOutputs();
            _totalLosses = 0;
            _totalPremium = 0;
            _oldSoldCount = soldCount;
            soldCount = 0;
            _renewCount = 0;
            _serviceCount = 0;
            _badServicecount = 0;
            _totalCustomerLocation = 0;
            for(int i=0;i<_customerYears.Length; i++)
            {
                _customerYears[i] = _customerBought[i] ? _customerYears[i] + 1 : 0;
                _customerBought[i] = false;
            }
        }
        internal void Close()
        {
            outputs.Close();
        }
        internal int SetNewRank(int newRank)
        {
            int oldRank = _rank;
            _rank = newRank;
            return _rank == oldRank ? 0 : 1;
        }
        internal double CalculatePremium(double marketPremium)
        {
            _newPremium = marketPremium + _serviceSpendPerCustomer;
            return _newPremium;
        }
        internal double CalculateSpending(double[] lastServiceSpendPerCustomer, double[] lastProfit)
        {
            _serviceSpendPerCustomer = _spendStrategy.GetServiceSpendPerCustomer(lastServiceSpendPerCustomer, lastProfit);
            _goodServiceProb = Math.Max(0,Math.Min(1.0,_serviceSpendPerCustomer/_maxServiceSpendPerCustomer));
            return _serviceSpendPerCustomer;
        }
        internal int CalculateServiceExperience()
        {
            bool goodService = _localRand.NextDouble() <= _goodServiceProb;
            _serviceCount++;
            if (goodService == false) { _badServicecount++; }     
            return goodService ? 1 : -1;
        }
        internal double GetRenewalPrice(int customerIndex)
        {
            return _newPremium * Math.Pow(1 + _renewMarkup, (double)Math.Min(_customerYears[customerIndex], _renewMarkupMaxTerm));

        }
        internal void AddCustomer(int customerIndex, double customerLocation)
        {
            bool renewal = _customerYears[customerIndex] != 0;
            double premium = renewal ? GetRenewalPrice(customerIndex) : _newPremium;
           soldCount++;
            if (renewal) { _renewCount++; }
            _totalPremium += premium;
            _totalCustomerLocation += customerLocation;
            _customerBought[customerIndex] = true;
        }
        internal void AddLoss(double Amount)
        {
            _totalLosses += Amount;
        }
        internal double CalculateProfit()
        {
            _totalProfit = _totalPremium - _totalLosses - _serviceSpendPerCustomer * soldCount;
            return _totalProfit;
        }
    }
}
