﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.IO;
using CsvHelper;

namespace CustomerBehaviour
{
    class Outputs
    {
        readonly string _outputFolder;
        readonly List<WriterPair> outputList;
        internal Outputs(string outputFolder)
        {
            _outputFolder = outputFolder;
            outputList = new List<WriterPair>();
        }
        internal Outputs(string outputFolder, string[] outputNames)
        {
            _outputFolder = outputFolder;
            outputList = new List<WriterPair>();
            int numNames = outputNames.Length;
            for(int i=0;i< numNames;i++)
            {
                outputList.Add(new WriterPair(outputFolder, outputNames[i]));
            }
        }
        internal void AddOutput(string outputName)
        {
            outputList.Add(new WriterPair(_outputFolder, outputName));
        }
        internal void SetValue(string outputName, double value)
        {
            outputList.Find(x => x.name == outputName).value = value;
        }
        internal void SetValues(double[] values)
        {
            for (int i = 0; i < outputList.Count; i++)
            {
                outputList[i].value=values[i];
            }
        }
        internal double GetValue(string outputName)
        {
            return outputList.Find(x => x.name == outputName).value;
        }
        internal void WriteOutputs()
        {
            for (int i = 0; i <outputList.Count; i++)
            {
                outputList[i].cw.WriteField(outputList[i].value);
            }
        }
        internal void NextSim()
        {
            for (int i = 0; i < outputList.Count; i++)
            {
                outputList[i].cw.NextRecord();
            }
        }
        internal void Close()
        {
            for (int i = 0; i < outputList.Count; i++)
            {
                outputList[i].sw.Close();
            }
        }
    }

    class CustomerOutputs
    {
        readonly string _outputFolder;
        readonly int numInsurers;
        readonly int numCustomers;
        readonly bool outputDistributions;
        readonly List<WriterPair> outputList;
        internal CustomerOutputs(string outputFolder, int insurerCount, int customerCount, bool writeDistributions)
        {
            _outputFolder = outputFolder;
            outputList = new List<WriterPair>();
            numInsurers = insurerCount;
            numCustomers = customerCount;
            outputDistributions = writeDistributions;
        }
        internal void AddOutput(string outputName)
        {
            for (int i = 0; i < numInsurers; i++)
            {
                outputList.Add(new WriterPair(_outputFolder, "Insurer " + i + "_" + outputName + "_Mean"));
                if (outputDistributions)
                {
                    for (int j = 0; j <= 100; j++)
                    {
                        outputList.Add(new WriterPair(_outputFolder, "Insurer " + i + "_" + outputName + "_" + (j)));
                    }
                }
            }
        }

        internal void SetValues(List<List<double[]>> customerValuesByInsurer)
        {
            int countOutList = 0;
            for (int valCount = 0; valCount < customerValuesByInsurer.Count; valCount++)
            {
                for (int i = 0; i < numInsurers; i++)
                {
                    //first get the required info for the insurer from the by-customer outputs
                    double[] insurerValuesByCustomer = new double[numCustomers];
                    for(int j=0;j<numCustomers;j++)
                    { 
                        insurerValuesByCustomer[j]=customerValuesByInsurer[valCount][j][i];
                    }
                    //Now set the mean output value
                    outputList[countOutList].value=insurerValuesByCustomer.Sum()/(double)numCustomers;
                    countOutList++;
                    //Now for the percentiles
                    if (outputDistributions)
                    {
                        Array.Sort(insurerValuesByCustomer);
                        for (int j = 0; j <= 100; j++)
                        {
                            double exactIndex = (numCustomers - 1) * 0.01 * j;
                            int lowIndex = (int)exactIndex;
                            double lowPerc = insurerValuesByCustomer[lowIndex];
                            double highPerc = lowIndex == numCustomers - 1 ? lowPerc : insurerValuesByCustomer[lowIndex + 1];
                            exactIndex -= lowIndex;
                            outputList[countOutList].value = exactIndex * highPerc + (1 - exactIndex) * lowPerc;
                            countOutList++;
                        }
                    }
                }
            }
        }
        internal void WriteOutputs()
        {
            for (int i = 0; i < outputList.Count; i++)
            {
                outputList[i].cw.WriteField(outputList[i].value);
            }
        }
        internal void NextSim()
        {
            for (int i = 0; i < outputList.Count; i++)
            {
                outputList[i].cw.NextRecord();
            }
        }
        internal void Close()
        {
            for (int i = 0; i < outputList.Count; i++)
            {
                outputList[i].sw.Close();
            }
        }
    }

    class WriterPair
    {
        internal string name;
        internal double value;
        internal StreamWriter sw;
        internal CsvWriter cw;
        internal WriterPair(string outputFolder, string fileName)
        {
            name = fileName;
            sw= new StreamWriter(outputFolder + "\\" + fileName + ".csv");
            cw = new CsvWriter(sw, System.Globalization.CultureInfo.CurrentCulture);
        }
    }
}
