﻿using System;
using System.IO;
using CsvHelper;

namespace CustomerBehaviour
{
    class Program
    {
        static void Main(string[] args)
        {
            string parentFolder = System.IO.Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly().Location);
            string configFile = Path.GetFullPath(Path.Combine(parentFolder, "..\\..\\..\\..\\data\\config.csv"));
            string resultsFolder = Path.GetFullPath(Path.Combine(parentFolder, "..\\..\\..\\..\\results\\"));
            if (System.IO.Directory.Exists(resultsFolder) == false) { System.IO.Directory.CreateDirectory(resultsFolder); }
            if (File.Exists(configFile))
            {
                StreamReader SReadSim = new StreamReader(configFile);
                CsvReader CSVReadSim = new CsvReader(SReadSim, System.Globalization.CultureInfo.CurrentCulture);
                CSVReadSim.Read();
                while (CSVReadSim.Read())
                {
                    string writeDistStr = CSVReadSim[29].ToLower().Substring(0, 1);
                    bool writeDists = writeDistStr == "t" || writeDistStr == "y";
                    string hetStr = CSVReadSim[18].ToLower().Substring(0, 1);
                    bool heterogeneity = hetStr == "t" || hetStr == "y";
                    RunSims(Path.Combine(resultsFolder,CSVReadSim[0]),
                        Convert.ToInt32(CSVReadSim[1]), Convert.ToInt32(CSVReadSim[2]), Convert.ToInt32(CSVReadSim[3]),
                        Convert.ToDouble(CSVReadSim[4]), Convert.ToDouble(CSVReadSim[5]), Convert.ToDouble(CSVReadSim[6]),
                        Convert.ToDouble(CSVReadSim[7]), Convert.ToDouble(CSVReadSim[8]), Convert.ToDouble(CSVReadSim[9]),
                        Convert.ToDouble(CSVReadSim[10]), Convert.ToInt32(CSVReadSim[11]), Convert.ToInt32(CSVReadSim[12]),
                        Convert.ToInt32(CSVReadSim[13]), Convert.ToInt32(CSVReadSim[14]), Convert.ToDouble(CSVReadSim[15]),
                        Convert.ToDouble(CSVReadSim[16]), Convert.ToDouble(CSVReadSim[17]), heterogeneity, 
                        Convert.ToDouble(CSVReadSim[19]), Convert.ToDouble(CSVReadSim[20]), Convert.ToDouble(CSVReadSim[21]), 
                        Convert.ToDouble(CSVReadSim[22]), Convert.ToDouble(CSVReadSim[23]), Convert.ToDouble(CSVReadSim[24]), 
                        Convert.ToDouble(CSVReadSim[25]), Convert.ToDouble(CSVReadSim[26]), Convert.ToDouble(CSVReadSim[27]), 
                        Convert.ToDouble(CSVReadSim[28]), writeDists
                        );
                }
                CSVReadSim.Dispose();
                SReadSim.Close();
            }
        }

        static void RunSims(string outputFolder, int numSims, int numTimesteps, int startSeed, double oldPremium, double oldOldPremium, double premiumAlpha0, double premiumAlpha1, double premiumAlpha2, double marketPremiumStdDev, double renewPremiumMarkup, int renewMarkupMaxTerm, int numCustomers, int numInsurers, int meanNetworkDegree, double networkRewireProb, double maxServiceSpendPerCustomer, double constServiceSpendPerCustomer, bool heterogeneity, double distanceMultiplier, double serviceMultiplier, double memoryFactor, double influenceFactor, double renewAlpha, double renewBeta, double freqLambda, double sevAlpha, double sevBeta, double transmissionProb, bool writeDistributions)
        {
            if (System.IO.Directory.Exists(outputFolder)) { System.IO.Directory.Delete(outputFolder, true); }
            System.IO.Directory.CreateDirectory(outputFolder);
            Random runRand = new Random(startSeed);
            ModelObject modelObj = new ModelObject(runRand.Next(), outputFolder, oldPremium, oldOldPremium, premiumAlpha0, premiumAlpha1, premiumAlpha2, marketPremiumStdDev, renewPremiumMarkup, renewMarkupMaxTerm, numCustomers, numInsurers, meanNetworkDegree, networkRewireProb, maxServiceSpendPerCustomer, MakeConstantStrategyList(numInsurers, constServiceSpendPerCustomer, maxServiceSpendPerCustomer, heterogeneity, runRand), distanceMultiplier, serviceMultiplier, memoryFactor, influenceFactor, renewAlpha, renewBeta, freqLambda, sevAlpha, sevBeta, transmissionProb, writeDistributions);
            for (int sim = 0; sim < numSims; sim++)
            {
                for (int timeStep = 0; timeStep < numTimesteps; timeStep++)
                {
                    modelObj.RunTimeStep();
                }
                modelObj.NextSim(MakeConstantStrategyList(numInsurers, constServiceSpendPerCustomer, maxServiceSpendPerCustomer,heterogeneity, runRand));
            }
            modelObj.Close();
        }
        static ISpendStrategy[] MakeConstantStrategyList(int numInsurers, double constServiceSpendPerCustomer, double maxServiceSpendPerCustomer, bool heterogeneity, Random hetRand)
        {
            int hetNums = (int)(maxServiceSpendPerCustomer / constServiceSpendPerCustomer);
            int decCount = BitConverter.GetBytes(decimal.GetBits((decimal)constServiceSpendPerCustomer)[3])[2];
            ISpendStrategy[] insStrategyList = new ISpendStrategy[numInsurers];
            for (int i = 0; i < numInsurers; i++)
            {
                insStrategyList[i] = new ConstSpend(heterogeneity ? Math.Round(constServiceSpendPerCustomer*hetRand.Next(hetNums+1), decCount) : constServiceSpendPerCustomer);
            }
            return insStrategyList;
        }
    }
}
